import logging
from threading import Semaphore

from qasync import asyncSlot
from qcanvas_api_clients.canvas import CanvasClient, CanvasClientConfig
from qcanvas_api_clients.panopto import PanoptoClient, PanoptoClientConfig
from qcanvas_api_clients.util.request_exceptions import ConfigInvalidError
from qtpy.QtCore import QUrl, Signal, Slot
from qtpy.QtGui import QDesktopServices, QIcon
from qtpy.QtWidgets import *

import qcanvas.util.settings as settings
from qcanvas import icons
from qcanvas.util import is_url
from qcanvas.util.layouts import grid_layout_widget, layout

_logger = logging.getLogger(__name__)

_tutorial_url = (
    "https://www.iorad.com/player/2053777/Canvas---How-to-generate-an-access-token-"
)


class SetupDialog(QDialog):
    closed = Signal()

    def __init__(self):
        super().__init__()

        self.setWindowTitle("Configure QCanvas")
        self.setMinimumSize(550, 200)
        self.resize(550, 200)
        self.setWindowIcon(QIcon(icons.main_icon))

        self._semaphore = Semaphore()
        self._canvas_url_box = QLineEdit(settings.client.canvas_url)
        self._canvas_api_key_box = QLineEdit(settings.client.canvas_api_key)
        self._canvas_api_key_box.setEchoMode(QLineEdit.EchoMode.Password)
        self._panopto_url_box = QLineEdit(settings.client.panopto_url)
        self._button_box = self._setup_button_box()
        self._button_box.accepted.connect(self._accepted)
        self._button_box.helpRequested.connect(self._help_requested)
        self._waiting_indicator = self._setup_progress_bar()

        self.setLayout(
            layout(
                QVBoxLayout,
                grid_layout_widget(
                    [
                        [QLabel("Canvas URL"), self._canvas_url_box],
                        [QLabel("Canvas API Key"), self._canvas_api_key_box],
                        [QLabel("Panopto URL"), self._panopto_url_box],
                    ]
                ),
                self._waiting_indicator,
                self._button_box,
            )
        )

    def _setup_button_box(self) -> QDialogButtonBox:
        box = QDialogButtonBox()
        box.addButton(QDialogButtonBox.StandardButton.Ok)
        box.addButton("Get a canvas API key", QDialogButtonBox.ButtonRole.HelpRole)
        return box

    def _setup_progress_bar(self) -> QProgressBar:
        progress = QProgressBar()
        progress.setMaximum(0)
        progress.setMinimum(0)
        self._set_retain_size_when_hidden(progress)
        progress.hide()
        return progress

    def _set_retain_size_when_hidden(self, widget: QWidget) -> None:
        size_policy = widget.sizePolicy()
        size_policy.setRetainSizeWhenHidden(True)
        widget.setSizePolicy(size_policy)

    @asyncSlot()
    async def _accepted(self) -> None:
        if self._semaphore.acquire(False):
            try:
                self._clear_errors()
                if self._check_inputs():
                    return

                self._waiting_indicator.setVisible(True)

                canvas_config = CanvasClientConfig(
                    api_token=self._canvas_api_key_box.text().strip(),
                    canvas_url=self._get_url(self._canvas_url_box),
                )

                if not await self._check_canvas_config(canvas_config):
                    return

                if not await self._check_panopto_config(canvas_config):
                    self._show_panopto_help()
                    return

            finally:
                self._waiting_indicator.setVisible(False)
                self._semaphore.release()

            _logger.debug("Credentials are a-ok!")
            self._save_and_close()
        else:
            _logger.debug("Validation already in progress")

    def _clear_errors(self) -> None:
        for line_edit in [
            self._canvas_url_box,
            self._panopto_url_box,
            self._canvas_api_key_box,
        ]:
            line_edit.setStyleSheet(None)
            line_edit.setToolTip(None)

    def _check_inputs(self) -> bool:
        had_error = False

        if not is_url(self._get_url(self._canvas_url_box)):
            had_error = True
            self._show_error(self._canvas_url_box, "Canvas URL is invalid")
        if len(self._canvas_api_key_box.text().strip()) == 0:
            had_error = True
            self._show_error(self._canvas_api_key_box, "Canvas API key is empty")
        if not is_url(self._get_url(self._panopto_url_box)):
            had_error = True
            self._show_error(self._panopto_url_box, "Panopto URL is invalid")

        return had_error

    def _get_url(self, line_edit: QLineEdit) -> str:
        url = line_edit.text().strip()

        if not url.startswith("http"):
            return "https://" + url
        else:
            return url

    async def _check_canvas_config(self, canvas_config: CanvasClientConfig) -> bool:
        try:
            await CanvasClient.verify_config(canvas_config)
            return True
        except ConfigInvalidError:
            self._show_error(self._canvas_api_key_box, "Canvas API key is invalid")
            return False

    def _show_error(self, line_edit: QLineEdit, text: str) -> None:
        line_edit.setToolTip(text)
        self._waiting_indicator.hide()
        self._highlight_line_edit(line_edit)

    def _highlight_line_edit(self, line_edit: QLineEdit) -> None:
        line_edit.setStyleSheet("QLineEdit { border: 1px solid red }")

    async def _check_panopto_config(self, canvas_config: CanvasClientConfig) -> bool:
        client = CanvasClient(canvas_config)
        try:
            await PanoptoClient.verify_config(
                PanoptoClientConfig(panopto_url=self._get_url(self._panopto_url_box)),
                client,
            )
            return True
        except ConfigInvalidError:
            return False
        finally:
            await client.aclose()

    def _show_panopto_help(self):
        msg = QMessageBox(
            QMessageBox.Icon.Information,
            "Panopto Authentication",
            "In order for QCanvas to use panopto, you need to link your panopto account to your canvas account. "
            "A page will open in your web browser to do this when you click OK. It may ask you to sign into canvas.\n\n"
            'Please tick "Remember my authorisation for this service" or QCanvas may not function correctly.\n\n'
            "QCanvas can't access anything entered in your browser.",
            QMessageBox.StandardButton.Ok,
            self,
        )
        msg.accepted.connect(self._open_panopto_login)
        msg.show()

    @Slot()
    def _open_panopto_login(self) -> None:
        url = QUrl(self._get_url(self._panopto_url_box))
        url.setPath("/Panopto/Pages/Auth/Login.aspx")
        url.setQuery("instance=Canvas&AllowBounce=true")
        QDesktopServices.openUrl(url)

    def _save_and_close(self) -> None:
        settings.client.canvas_url = self._get_url(self._canvas_url_box)
        settings.client.panopto_url = self._get_url(self._panopto_url_box)
        settings.client.canvas_api_key = self._canvas_api_key_box.text().strip()
        self.closed.emit()
        self.close()

    @Slot()
    def _help_requested(self) -> None:
        msg = QMessageBox(
            QMessageBox.Icon.Information,
            "Help",
            "An interactive tutorial will open in your browser when you click OK.\n\n"
            'Note that the "purpose" text doesn\'t matter and you can enter anything you want.\n\n'
            'You should also leave the "expires" item blank if you want the key to last forever.\n\n'
            "Don't share this key. You can revoke it at any time.",
            parent=self,
        )
        msg.accepted.connect(self._open_tutorial)
        msg.show()

    @Slot()
    def _open_tutorial(self) -> None:
        QDesktopServices.openUrl(QUrl(_tutorial_url))



import logging
from logging import DEBUG, INFO, WARNING

import qcanvas.app_start
from qcanvas.util import logs, paths

paths.data_storage().mkdir(parents=True, exist_ok=True)

logging.basicConfig(
    filemode="w",
    filename=paths.data_storage() / "debug.log",
    level="WARN",
)

logs.set_levels(
    {
        "qcanvas": INFO,
        "qcanvas.ui": WARNING,
        "qcanvas_backend": INFO,
        "qcanvas.ui.main_ui.status_bar_progress_display": DEBUG,
    }
)


def main():
    qcanvas.app_start.launch()


if __name__ == "__main__":
    main()



import logging
from typing import *

import qcanvas_backend.database.types as db
from qcanvas_backend.net.sync.sync_receipt import SyncReceipt
from qtpy.QtWidgets import *

from qcanvas.ui.course_viewer.content_tree import ContentTree
from qcanvas.ui.memory_tree import MemoryTreeWidgetItem

_logger = logging.getLogger(__name__)


class MailTree(ContentTree[db.Course]):
    @staticmethod
    def create_from_receipt(
        course: db.Course, *, sync_receipt: Optional[SyncReceipt]
    ) -> "MailTree":
        tree = MailTree(course.id)
        tree.reload(course, sync_receipt=sync_receipt)
        return tree

    def __init__(self, course_id: str):
        super().__init__(
            tree_name=f"course.{course_id}.mail",
            emit_selection_signal_for_type=db.CourseMessage,
        )
        self.ui_setup(
            header_text=["Subject", "Sender"],
            max_width=500,
            min_width=300,
            indentation=20,
        )

        self._adjust_header()

    def _adjust_header(self) -> None:
        header = self.header()
        header.setSectionResizeMode(0, QHeaderView.ResizeMode.Stretch)
        header.setSectionResizeMode(1, QHeaderView.ResizeMode.ResizeToContents)
        header.setStretchLastSection(False)

    def create_tree_items(
        self, course: db.Course, sync_receipt: Optional[SyncReceipt]
    ) -> Sequence[MemoryTreeWidgetItem]:
        widgets = []

        for message in course.messages:  # type: db.CourseMessage
            message_widget = self._create_mail_widget(message, sync_receipt)
            widgets.append(message_widget)

        return widgets

    def _create_mail_widget(
        self, message: db.CourseMessage, sync_receipt: Optional[SyncReceipt]
    ) -> MemoryTreeWidgetItem:
        message_widget = MemoryTreeWidgetItem(
            id=message.id,
            data=message,
            strings=[message.name, message.sender_name],
        )

        is_new = (
            sync_receipt is not None and message.id in sync_receipt.updated_messages
        )

        if is_new:
            self.mark_as_unseen(message_widget)

        return message_widget



import logging
from urllib.parse import urlparse

_logger = logging.getLogger(__name__)


def is_url(url: str) -> bool:
    # https://overflow.perennialte.ch/questions/7160737/how-to-validate-a-url-in-python-malformed-or-not#
    try:
        result = urlparse(url)
        return all([result.scheme, result.netloc])
    except ValueError:
        return False



import logging

import qdarktheme

_logger = logging.getLogger(__name__)

default_theme = "auto"


def ensure_theme_is_valid(theme: str) -> str:
    if theme not in ["auto", "light", "dark", "native"]:
        return default_theme
    else:
        return theme


def apply(theme: str) -> None:
    theme = ensure_theme_is_valid(theme)

    if theme != "native":
        qdarktheme.setup_theme(
            theme,
            custom_colors={"primary": "e02424"},
        )



import logging
from typing import Optional

import qcanvas_backend.database.types as db
from qcanvas_backend.net.resources.download.resource_manager import ResourceManager
from qcanvas_backend.net.sync.sync_receipt import SyncReceipt
from qtpy.QtWidgets import *

from qcanvas.ui.course_viewer.tabs.assignment_tab import AssignmentTab
from qcanvas.ui.course_viewer.tabs.mail_tab import MailTab
from qcanvas.ui.course_viewer.tabs.page_tab import PageTab
from qcanvas.util.basic_fonts import bold_font
from qcanvas.util.layouts import layout
from qcanvas.util.ui_tools import make_truncatable

_logger = logging.getLogger(__name__)


class CourseViewer(QWidget):
    def __init__(
        self,
        course: db.Course,
        downloader: ResourceManager,
        *,
        initial_sync_receipt: Optional[SyncReceipt] = None
    ):
        super().__init__()
        # todo this is a mess. there are several other messes like this too, do they all have to be a mess?
        self._course_label = QLabel(course.name)
        self._course_label.setFont(bold_font)
        make_truncatable(self._course_label)
        self._pages_tab = PageTab.create_from_receipt(
            course=course,
            downloader=downloader,
            sync_receipt=initial_sync_receipt,
        )
        self._assignments_tab = AssignmentTab.create_from_receipt(
            course=course,
            downloader=downloader,
            sync_receipt=initial_sync_receipt,
        )
        self._mail_tab = MailTab.create_from_receipt(
            course=course,
            downloader=downloader,
            sync_receipt=initial_sync_receipt,
        )
        self._tabs = QTabWidget()

        self._tabs.addTab(self._pages_tab, "Pages")
        self._tabs.addTab(self._assignments_tab, "Assignments")
        self._tabs.addTab(self._mail_tab, "Mail")
        # todo: move back to first when implemented
        self._tabs.addTab(QLabel("Not implemented"), "Files")
        # self._tabs.addTab(QLabel("Not implemented"), "Panopto") The meme lives on!

        self.setLayout(layout(QVBoxLayout, self._course_label, self._tabs))

    def reload(self, course: db.Course, *, sync_receipt: Optional[SyncReceipt]) -> None:
        # self._tabs.setTabText(1, "*Pages")
        self._pages_tab.reload(course, sync_receipt=sync_receipt)
        self._assignments_tab.reload(course, sync_receipt=sync_receipt)
        self._mail_tab.reload(course, sync_receipt=sync_receipt)



import logging
from threading import Semaphore
from typing import *

import qcanvas_backend.database.types as db
from qasync import asyncSlot
from qcanvas_backend.database.data_monolith import DataMonolith
from qcanvas_backend.net.sync.sync_receipt import SyncReceipt
from qcanvas_backend.qcanvas import QCanvas
from qtpy.QtCore import QUrl, Signal, Slot
from qtpy.QtGui import QDesktopServices, QIcon, QKeySequence
from qtpy.QtWidgets import *

from qcanvas import icons
from qcanvas.backend_connectors import FrontendResourceManager
from qcanvas.ui.course_viewer import CourseTree
from qcanvas.ui.main_ui.course_viewer_container import CourseViewerContainer
from qcanvas.ui.main_ui.options.quick_sync_option import QuickSyncOption
from qcanvas.ui.main_ui.options.sync_on_start_option import SyncOnStartOption
from qcanvas.ui.main_ui.options.theme_selection_menu import ThemeSelectionMenu
from qcanvas.ui.main_ui.status_bar_progress_display import StatusBarProgressDisplay
from qcanvas.util import paths, settings
from qcanvas.util.qurl_util import file_url
from qcanvas.util.ui_tools import create_qaction

_logger = logging.getLogger(__name__)


class QCanvasWindow(QMainWindow):
    _loaded = Signal()

    def __init__(self):
        super().__init__()

        self.setWindowTitle("QCanvas")
        self.setWindowIcon(QIcon(icons.main_icon))

        self._operation_semaphore = Semaphore()
        self._data: Optional[DataMonolith] = None
        self._qcanvas = QCanvas[FrontendResourceManager](
            canvas_config=settings.client.canvas_config,
            panopto_config=settings.client.panopto_config,
            storage_path=paths.data_storage(),
            resource_manager_class=FrontendResourceManager,
        )

        self._course_tree = CourseTree()
        self._course_tree.item_selected.connect(self._on_course_selected)
        self._course_tree.course_renamed.connect(self._on_course_renamed)
        self._sync_button = QPushButton("Synchronise")
        self._sync_button.clicked.connect(self._synchronise_requested)
        self._course_viewer_container = CourseViewerContainer(
            self._qcanvas.resource_manager
        )

        self.setCentralWidget(self._setup_main_layout())
        self.setStatusBar(StatusBarProgressDisplay())
        self._setup_menu_bar()
        self._restore_window_position()

        self._loaded.connect(self._on_app_loaded)
        self._loaded.emit()

    def _setup_menu_bar(self) -> None:
        menu_bar = self.menuBar()
        app_menu = menu_bar.addMenu("Actions")

        create_qaction(
            name="Synchronise",
            shortcut=QKeySequence("Ctrl+S"),
            triggered=self._synchronise_requested,
            parent=app_menu,
        )

        create_qaction(
            name="Open downloads folder",
            shortcut=QKeySequence("Ctrl+D"),
            triggered=self._open_downloads_folder,
            parent=app_menu,
        )

        create_qaction(
            name="Quick canvas login",
            shortcut=QKeySequence("Ctrl+O"),
            triggered=self._open_quick_auth_in_browser,
            parent=app_menu,
        )

        create_qaction(
            name="Mark all as seen", triggered=self._clear_new_items, parent=app_menu
        )

        create_qaction(
            name="Quit",
            shortcut=QKeySequence("Ctrl+Q"),
            triggered=lambda: self.close(),
            parent=app_menu,
        )

        options_menu = menu_bar.addMenu("Options")

        options_menu.addAction(QuickSyncOption(options_menu))
        options_menu.addAction(SyncOnStartOption(options_menu))
        options_menu.addMenu(ThemeSelectionMenu(self))

    def _restore_window_position(self):
        if settings.ui.last_geometry is not None:
            self.restoreGeometry(settings.ui.last_geometry)

        if settings.ui.last_window_state is not None:
            self.restoreState(settings.ui.last_window_state)

    def closeEvent(self, event):
        settings.ui.last_geometry = self.saveGeometry()
        settings.ui.last_window_state = self.saveState()

    def _setup_main_layout(self) -> QWidget:
        h_box = QHBoxLayout()

        h_box.addLayout(self._setup_course_column(), 1)
        h_box.addWidget(self._course_viewer_container, 5)

        widget = QWidget()
        widget.setLayout(h_box)
        return widget

    def _setup_course_column(self) -> QVBoxLayout:
        course_list_column = QVBoxLayout()
        course_list_column.addWidget(self._course_tree)
        course_list_column.addWidget(self._sync_button)

        return course_list_column

    @asyncSlot()
    async def _on_app_loaded(self) -> None:
        await self._qcanvas.init()
        self._course_tree.reload(await self._get_terms(), sync_receipt=None)

        if settings.client.sync_on_start:
            await self._synchronise()

    @asyncSlot()
    async def _synchronise_requested(self) -> None:
        await self._synchronise()

    async def _synchronise(self) -> None:
        if not self._operation_semaphore.acquire(False):
            _logger.debug("Sync operation already in progress")
            return

        try:
            # todo handle exceptions and PROGRESS!! better
            self._sync_button.setText("Sync in progress...")
            receipt = await self._qcanvas.synchronise_canvas(
                quick_sync=settings.client.quick_sync_enabled
            )
            await self._reload(receipt)
            self._sync_button.setText("Synchronise")
        except Exception as e:
            error = QErrorMessage(self)
            error.showMessage(str(e))

            raise e
        finally:
            self._operation_semaphore.release()

    async def _reload(self, receipt: Optional[SyncReceipt]) -> None:
        self._course_tree.reload(await self._get_terms(), sync_receipt=receipt)
        await self._course_viewer_container.reload_all(
            await self._get_courses(), sync_receipt=receipt
        )

    async def _get_terms(self) -> Sequence[db.Term]:
        return (await self._qcanvas.get_data()).terms

    async def _get_courses(self) -> Sequence[db.Course]:
        return (await self._qcanvas.get_data()).courses

    @Slot(db.Course)
    def _on_course_selected(self, course: Optional[db.Course]) -> None:
        if course is not None:
            self._course_viewer_container.load_course(course)
        else:
            self._course_viewer_container.show_blank()

    @asyncSlot(db.Course, str)
    async def _on_course_renamed(self, course: db.Course, new_name: str) -> None:
        _logger.debug("Rename %s -> %s", course.name, new_name)

        async with self._qcanvas.database.session() as session:
            session.add(course)
            course.configuration.nickname = new_name

    @asyncSlot()
    async def _open_quick_auth_in_browser(self) -> None:
        opening_progress_dialog = QProgressDialog("Opening canvas", None, 0, 0, self)
        opening_progress_dialog.setWindowTitle("Please wait")
        opening_progress_dialog.show()
        QDesktopServices.openUrl(
            QUrl(await self._qcanvas.canvas_client.get_temporary_session_url())
        )
        opening_progress_dialog.close()

    @Slot()
    def _open_downloads_folder(self) -> None:
        directory = self._qcanvas.resource_manager.downloads_folder
        directory.mkdir(parents=True, exist_ok=True)

        QDesktopServices.openUrl(file_url(directory))

    @asyncSlot()
    async def _clear_new_items(self) -> None:
        await self._reload(None)



import logging
from typing import Optional, Sequence

import qcanvas_backend.database.types as db
from qcanvas_backend.net.sync.sync_receipt import SyncReceipt
from qtpy.QtCore import Qt

from qcanvas.ui.course_viewer.content_tree import ContentTree
from qcanvas.ui.memory_tree import MemoryTreeWidgetItem

_logger = logging.getLogger(__name__)


class PageTree(ContentTree[db.Course]):
    @staticmethod
    def create_from_receipt(
        course: db.Course, *, sync_receipt: Optional[SyncReceipt]
    ) -> "PageTree":
        tree = PageTree(course.id)
        tree.reload(course, sync_receipt=sync_receipt)
        return tree

    def __init__(self, course_id: str):
        super().__init__(
            tree_name=f"course.{course_id}.modules",
            emit_selection_signal_for_type=db.ModulePage,
        )
        self.ui_setup(
            header_text="Content", indentation=15, max_width=300, min_width=150
        )

    def create_tree_items(
        self, course: db.Course, sync_receipt: Optional[SyncReceipt]
    ) -> Sequence[MemoryTreeWidgetItem]:
        widgets = []

        for module in course.modules:  # type: db.Module
            module_widget = self._create_module_widget(module, sync_receipt)
            widgets.append(module_widget)

            for page in module.pages:  # type: db.ModulePage
                page_widget = self._create_page_widget(page, sync_receipt)
                module_widget.addChild(page_widget)

        return widgets

    def _create_module_widget(
        self, module: db.Module, sync_receipt: Optional[SyncReceipt]
    ) -> MemoryTreeWidgetItem:
        module_widget = MemoryTreeWidgetItem(
            id=module.id, data=module, strings=[module.name]
        )
        module_widget.setFlags(Qt.ItemFlag.ItemIsEnabled)

        # todo add some helpers to SyncReceipt to make this less shit, and maybe use an empty syncreceipt instead of None
        is_new = sync_receipt is not None and module.id in sync_receipt.updated_modules

        if is_new:
            self.mark_as_unseen(module_widget)

        return module_widget

    def _create_page_widget(
        self, page: db.ModulePage, sync_receipt: Optional[SyncReceipt]
    ) -> MemoryTreeWidgetItem:
        page_widget = MemoryTreeWidgetItem(id=page.id, data=page, strings=[page.name])
        page_widget.setFlags(Qt.ItemFlag.ItemIsEnabled | Qt.ItemFlag.ItemIsSelectable)

        is_new = sync_receipt is not None and page.id in sync_receipt.updated_pages

        if is_new:
            self.mark_as_unseen(page_widget)

        return page_widget



import logging
from typing import *

import qcanvas_backend.database.types as db
from qcanvas_backend.net.resources.download.resource_manager import ResourceManager
from qcanvas_backend.net.sync.sync_receipt import SyncReceipt
from qtpy.QtCore import Qt
from qtpy.QtWidgets import *

from qcanvas.ui.course_viewer.course_viewer import CourseViewer

_logger = logging.getLogger(__name__)


class CourseViewerContainer(QStackedWidget):
    def __init__(self, downloader: ResourceManager):
        super().__init__()
        self._course_viewers: dict[str, CourseViewer] = {}
        self._downloader = downloader
        self._last_course_id: Optional[str] = None
        self._last_sync_receipt: Optional[SyncReceipt] = None
        self._placeholder = QLabel("No Course Selected")
        self._placeholder.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.addWidget(self._placeholder)

    def show_blank(self) -> None:
        self._last_course_id = None
        self.setCurrentWidget(self._placeholder)

    def load_course(self, course: db.Course) -> None:
        if course.id not in self._course_viewers:
            viewer = CourseViewer(
                course=course,
                downloader=self._downloader,
                initial_sync_receipt=self._last_sync_receipt,
            )
            self._course_viewers[course.id] = viewer
            self.addWidget(viewer)
        else:
            viewer = self._course_viewers[course.id]

        self.setCurrentWidget(viewer)
        self._last_course_id = course.id

    async def reload_all(
        self, courses: Sequence[db.Course], *, sync_receipt: Optional[SyncReceipt]
    ) -> None:
        self._last_sync_receipt = sync_receipt
        for course in courses:
            if course.id in self._course_viewers:
                viewer = self._course_viewers[course.id]
                viewer.reload(course, sync_receipt=sync_receipt)
